--[[TODO List
Scrollbar
Add ability to search for variables in the Global namespace instead of just specifying tables/functions

]]

Protractor = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0")
local Graph=AceLibrary("Graph-1.0")
local me=Protractor

me.ChartData={}
me.ProcessedData={}
me.consoleOptions = {
	type = 'group',
	args = {
		show={
			name = "Show",
			desc = "Shows the protractor window",
			type = 'execute',
			func = function() me.Frame:Show() end
		},
		reset={
			name = "Reset",
			desc = "Resets the current CPU Profiling info",
			type = 'execute',
			func = function()  ResetCPUUsage()  end
		},
		profiling={
			name="Profiling",
			desc="Is CPU Profiling Enabled (will ReloadUI if changed)",
			type='toggle',
			get=function() return GetCVar("scriptProfile")=="1" end,
			set=function(v)
				if v then
					SetCVar("scriptProfile", "1")
				else
					SetCVar("scriptProfile", "0")
				end
				ReloadUI()
			end,
			map = { [false] = "|cffff4040Disabled|r", [true] = "|cff40ff40Enabled|r" }
		},
	}
}

me:RegisterChatCommand({"/pro", "/protractor"}, me.consoleOptions)

function me:MemoryUsage()
	local count = GetNumAddOns()
	local i
	UpdateAddOnMemoryUsage()

	me:ResetChart()
	for i=1, count do
		local _, title, notes, _, _, _, _ = GetAddOnInfo(i)
		local memory = GetAddOnMemoryUsage(i)
		--me:Print(title.." uses "..memory.."kb")
		me:AddChartData(title,memory)
	end
	me:CompleteChart()

	me.Frame.Title:SetText("Memory Usage")
	me.Frame.UsageTitle:SetText("Memory (kb)")
end

function me:CPUUsage()
	if GetCVar("scriptProfile")~="1" then
		Protractor:Print("CPU Profiling is currently disabled")
		return
	end

	local count = GetNumAddOns()
	UpdateAddOnCPUUsage()

	me:ResetChart()

	for i=1, count do
		local _, title, notes, _, _, _, _ = GetAddOnInfo(i)
		local time = GetAddOnCPUUsage(i)

		me:AddChartData(title,time)
		if time>0 then
			--me:Print(title.." uses "..time.."s")
		end
	end
	me:CompleteChart()

	me.Frame.Title:SetText("CPU Usage")
	me.Frame.UsageTitle:SetText("Time (ms)")
end

function me:ResetChart()
	for k,v in pairs(me.ChartData) do
		me:FreeTable(v)
		me.ChartData[k]=nil
	end
	me.ChartTotal=0
end

local FreeTables={}
function me:GetTable()
	local t
	if #FreeTables>0 then
		t=FreeTables[1]
		tremove(FreeTables,1)
		return t
	end
	return {}
end

function me:FreeTable(t)
	for v in pairs(FreeTables) do
		if t==v then
			return
		end
	end
	FreeTables[#FreeTables+1]=t
end


function me:AddChartData(name,amount,count)
	me.ChartTotal=me.ChartTotal+amount

	local Pos=#me.ChartData+1
	me.ChartData[Pos]=me:GetTable()

	me.ChartData[Pos].name=name
	me.ChartData[Pos].amount=amount
	me.ChartData[Pos].count=count
end

function SortChart(a,b)
	if a.amount>b.amount then
		return true
	end
end

function me:SelectingPiece(piece)
	local offset = FauxScrollFrame_GetOffset(me.Frame.ScrollBar)
	for i=1,10 do
		me.Frame.RowTexture[i]:Hide()
	end

	if piece then
		piece=piece-offset

		if piece<=10 and piece>=1 then
			me.Frame.RowTexture[piece]:Show()
		end
	end
end



function me:CompleteChart()
	local CollectedData=me.ChartData
	local Cutoff, Total,Pos

	me.LeftOver=0
	Total=me.ChartTotal
	Cutoff=Total*me.CutoffValue

	for _, v in pairs(CollectedData) do
		if v.amount<Cutoff then
			me.LeftOver=me.LeftOver+v.amount
		else
			Pos=#me.ProcessedData+1
			me.ProcessedData[Pos]=v
			me.ProcessedData[Pos].percent=100*v.amount/Total
		end
	end

	table.sort(me.ProcessedData,SortChart)

	

	local CurColor=0

	if me.PieChart then
		me.PieChart:ResetPie()
	else
		me:CreateFrame()
	end
	local g=me.PieChart
	
	CurColor=1
	for _, v in pairs(me.ProcessedData) do
		CurColor=g:AddPie(v.percent)
		v.Color=CurColor
	end
	
	me.OtherColor=g:CompletePie()
	
	me:UpdateWindow()

	for k in pairs(me.ProcessedData) do
		me.ProcessedData[k]=nil
	end
end

function me:UpdateWindow()
	FauxScrollFrame_Update(me.Frame.ScrollBar, table.getn(me.ProcessedData)+1, 10, 14)
	local offset = FauxScrollFrame_GetOffset(me.Frame.ScrollBar)

	local AddedLeftOver=false
	local AnyCounts=false
	for i=1,10 do
		local CurEntry=me.ProcessedData[i+offset]
		if CurEntry then
			me.Frame.Name[i]:SetText(CurEntry.name)
			
			--Does it fit?
			if me.Frame.Name[i]:GetStringWidth()>180 then
				local NameFrame=me.Frame.Name[i]
				local NameString=CurEntry.name

				repeat
					NameString=string.sub(NameString,1,string.len(NameString)-1)
					NameFrame:SetText(NameString.."...")
				until NameFrame:GetStringWidth()<=180
			end


			me.Frame.Usage[i]:SetText(math.floor(CurEntry.amount+0.5))
			me.Frame.Percent[i]:SetText(math.floor(10*CurEntry.percent)/10 .. "%")
			me.Frame.Key[i]:SetVertexColor(CurEntry.Color[1],CurEntry.Color[2],CurEntry.Color[3])
			me.Frame.RowTexture[i]:SetVertexColor((CurEntry.Color[1]+0.5)/2,(CurEntry.Color[2]+0.5)/2,(CurEntry.Color[3]+0.5)/2,0.75)

			if CurEntry.count then
				me.Frame.Count[i]:SetText(CurEntry.count)
				me.Frame.Count[i]:Show()
				AnyCounts=true
			else
				me.Frame.Count[i]:Hide()
			end
			
			me.Frame.Name[i]:Show()
			me.Frame.Usage[i]:Show()
			me.Frame.Percent[i]:Show()
			me.Frame.Key[i]:Show()
		elseif not AddedLeftOver then
			me.Frame.Name[i]:SetText("Others")
			me.Frame.Usage[i]:SetText(math.floor(me.LeftOver+0.5))
			me.Frame.Percent[i]:SetText(math.floor(1000*me.LeftOver/me.ChartTotal)/10 .. "%")
			me.Frame.Key[i]:SetVertexColor(me.OtherColor[1],me.OtherColor[2],me.OtherColor[3])
			me.Frame.RowTexture[i]:SetVertexColor((me.OtherColor[1]+0.5)/2,(me.OtherColor[2]+0.5)/2,(me.OtherColor[3]+0.5)/2,0.75)

			me.Frame.Name[i]:Show()
			me.Frame.Usage[i]:Show()
			me.Frame.Percent[i]:Show()
			me.Frame.Key[i]:Show()
			me.Frame.Count[i]:Hide()

			AddedLeftOver=true
		else
			me.Frame.Name[i]:Hide()
			me.Frame.Usage[i]:Hide()
			me.Frame.Percent[i]:Hide()
			me.Frame.Key[i]:Hide()
			me.Frame.Count[i]:Hide()
		end
	end

	if AnyCounts then
		me.Frame.CountTitle:Show()
	else
		me.Frame.CountTitle:Hide()
	end
end

function me:NextDisplay()
	me.Displaying=me.Displaying+1

	if me.Displaying>=me.MaxDisplay then
		me.Displaying=0
	end

	me.DisplayFunc[me.Displaying]()
end

function me:PrevDisplay()
	me.Displaying=me.Displaying-1

	if me.Displaying<0 then
		me.Displaying=me.MaxDisplay-1
	end

	me.DisplayFunc[me.Displaying]()
end

function me:CreateFrame()
	me.Frame=CreateFrame("Frame","Protractor",UIParent)

	local theFrame=me.Frame

	theFrame:ClearAllPoints()
	theFrame:SetPoint("CENTER",UIParent)

	theFrame:SetWidth(570)
	theFrame:SetHeight(210)
	theFrame:SetBackdrop({bgFile = "Interface/Tooltips/UI-Tooltip-Background", 
                                            edgeFile = "Interface/Tooltips/UI-Tooltip-Border", 
                                            tile = true, tileSize = 16, edgeSize = 16, 
                                            insets = { left = 4, right = 4, top = 4, bottom = 4 }});

	theFrame:SetBackdropColor(0,0,0,0.5);

	theFrame:SetScript("OnMouseDown", function() 
						if ( ( ( not me.Frame.isLocked ) or ( me.Frame.isLocked == 0 ) ) and ( arg1 == "LeftButton" ) ) then
						  me.Frame:StartMoving();
						  me.Frame.isMoving = true;
						 end
						end)
	theFrame:SetScript("OnMouseUp", function() 
						if ( me.Frame.isMoving ) then
						  me.Frame:StopMovingOrSizing();
						  me.Frame.isMoving = false;
						 end
						end)
					
	theFrame:SetScript("OnHide", function() 
						if ( me.Frame.isMoving ) then
						  me.Frame:StopMovingOrSizing();
						  me.Frame.isMoving = false;
						 end
						end)
	

	theFrame:SetMovable(true)
	theFrame:EnableMouse()
	theFrame:Show()

	me.Frame.Title=theFrame:CreateFontString(nil,"OVERLAY","GameFontNormal")
	me.Frame.Title:SetPoint("TOP",theFrame,"TOP",0,-8)
	me.Frame.Title:SetTextColor(1.0,1.0,1.0)

	me.Frame.LeftButton=CreateFrame("Button",nil,theFrame)
	me.Frame.LeftButton:SetNormalTexture("Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Up.blp")
	me.Frame.LeftButton:SetPushedTexture("Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Down.blp")
	me.Frame.LeftButton:SetWidth(18)
	me.Frame.LeftButton:SetHeight(18)
	me.Frame.LeftButton:SetPoint("TOP",theFrame,"TOP",-80,-5)
	me.Frame.LeftButton:SetScript("OnClick",function() 
							me:PrevDisplay()
						end)

	me.Frame.RightButton=CreateFrame("Button",nil,theFrame)
	me.Frame.RightButton:SetNormalTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Up.blp")
	me.Frame.RightButton:SetPushedTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Down.blp")
	
	me.Frame.RightButton:SetWidth(18)
	me.Frame.RightButton:SetHeight(18)
	me.Frame.RightButton:SetPoint("TOP",theFrame,"TOP",80,-5)
	me.Frame.RightButton:SetScript("OnClick",function() 
							me:NextDisplay()
						end)

	me.Frame.CloseButton=CreateFrame("Button",nil,theFrame)
	me.Frame.CloseButton:SetNormalTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Up.blp")
	me.Frame.CloseButton:SetPushedTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Down.blp")
	me.Frame.CloseButton:SetHighlightTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Highlight.blp")
	me.Frame.CloseButton:SetWidth(24)
	me.Frame.CloseButton:SetHeight(24)
	me.Frame.CloseButton:SetPoint("TOPRIGHT",theFrame,"TOPRIGHT",-2,-2)
	me.Frame.CloseButton:SetScript("OnClick",function() me.Frame:Hide() end)

	me.Frame.RefreshButton=CreateFrame("Button",nil,me.Frame)
	me.Frame.RefreshButton:SetNormalTexture("Interface\\Buttons\\UI-RotationLeft-Button-Up.blp")
	me.Frame.RefreshButton:SetPushedTexture("Interface\\Buttons\\UI-RotationLeft-Button-Down.blp")
	me.Frame.RefreshButton:SetWidth(20)
	me.Frame.RefreshButton:SetHeight(20)
	me.Frame.RefreshButton:SetPoint("RIGHT",me.Frame.CloseButton,"LEFT",0,0)
	me.Frame.RefreshButton:SetScript("OnClick",function() me.DisplayFunc[me.Displaying]() end)

	

	me.Frame.NameTitle=theFrame:CreateFontString(nil,"OVERLAY","GameFontNormal")
	me.Frame.NameTitle:SetPoint("TOP",theFrame,"TOP",-55,-21)
	me.Frame.NameTitle:SetTextColor(1.0,1.0,1.0)
	me.Frame.NameTitle:SetText("Name")
	
	me.Frame.KeyTitle=theFrame:CreateFontString(nil,"OVERLAY","GameFontNormal")
	me.Frame.KeyTitle:SetPoint("RIGHT",me.Frame.NameTitle,"LEFT",-4,0)
	me.Frame.KeyTitle:SetTextColor(1.0,1.0,1.0)
	me.Frame.KeyTitle:SetText("K")


	me.Frame.UsageTitle=theFrame:CreateFontString(nil,"OVERLAY","GameFontNormal")
	me.Frame.UsageTitle:SetPoint("CENTER",me.Frame.NameTitle,"LEFT",200,0)
	me.Frame.UsageTitle:SetTextColor(1.0,1.0,1.0)
	me.Frame.UsageTitle:SetText("Usage")

	me.Frame.PercentTitle=theFrame:CreateFontString(nil,"OVERLAY","GameFontNormal")
	me.Frame.PercentTitle:SetPoint("CENTER",me.Frame.UsageTitle,"CENTER",50,0)
	me.Frame.PercentTitle:SetTextColor(1.0,1.0,1.0)
	me.Frame.PercentTitle:SetText("%")

	me.Frame.CountTitle=theFrame:CreateFontString(nil,"OVERLAY","GameFontNormal")
	me.Frame.CountTitle:SetPoint("CENTER",me.Frame.PercentTitle,"CENTER",50,0)
	me.Frame.CountTitle:SetTextColor(1.0,1.0,1.0)
	me.Frame.CountTitle:SetText("Count")
	me.Frame.CountTitle:Hide()

	me.Frame.Key={}
	me.Frame.Name={}
	me.Frame.Usage={}
	me.Frame.Percent={}
	me.Frame.Count={}
	me.Frame.RowTexture={}
	for i=1,10 do
		me.Frame.Key[i]=theFrame:CreateTexture(nil,"OVERLAY")
		me.Frame.Key[i]:SetPoint("TOP",me.Frame.KeyTitle,"TOP",0,-14*i)
		me.Frame.Key[i]:SetTexture("Interface\\Buttons\\WHITE8X8.blp")
		me.Frame.Key[i]:SetWidth(12)
		me.Frame.Key[i]:SetHeight(12)

		--me.Frame.Key[i]:Hide()

		me.Frame.Name[i]=theFrame:CreateFontString(nil,"OVERLAY","GameFontNormal")
		me.Frame.Name[i]:SetPoint("TOPLEFT",me.Frame.NameTitle,"TOPLEFT",0,-14*i)
		me.Frame.Name[i]:SetTextColor(1.0,1.0,1.0)
		me.Frame.Name[i]:SetText("")

		me.Frame.Usage[i]=theFrame:CreateFontString(nil,"OVERLAY","GameFontNormal")
		me.Frame.Usage[i]:SetPoint("TOP",me.Frame.UsageTitle,"TOP",0,-14*i)
		me.Frame.Usage[i]:SetTextColor(1.0,1.0,1.0)
		me.Frame.Usage[i]:SetText("")

		me.Frame.Percent[i]=theFrame:CreateFontString(nil,"OVERLAY","GameFontNormal")
		me.Frame.Percent[i]:SetPoint("TOP",me.Frame.PercentTitle,"TOP",0,-14*i)
		me.Frame.Percent[i]:SetTextColor(1.0,1.0,1.0)
		me.Frame.Percent[i]:SetText("")

		me.Frame.Count[i]=theFrame:CreateFontString(nil,"OVERLAY","GameFontNormal")
		me.Frame.Count[i]:SetPoint("TOP",me.Frame.CountTitle,"TOP",0,-14*i)
		me.Frame.Count[i]:SetTextColor(1.0,1.0,1.0)
		me.Frame.Count[i]:SetText("")

		me.Frame.RowTexture[i]=theFrame:CreateTexture(nil,"BACKGROUND")
		me.Frame.RowTexture[i]:SetPoint("TOPLEFT",me.Frame.Name[i],"TOPLEFT",0,0)
		me.Frame.RowTexture[i]:SetPoint("BOTTOMRIGHT",me.Frame.Percent[i],"BOTTOM",80,0)
		me.Frame.RowTexture[i]:SetTexture("Interface\\Buttons\\WHITE8X8.blp")
		me.Frame.RowTexture[i]:Hide()
	end

	me.PieChart=Graph:CreateGraphPieChart("Protractor_PieChart",theFrame,"LEFT","LEFT",0,-5,200,200)
	me.PieChart:SetSelectionFunc(me.SelectingPiece)

	me.Frame.TextBox=CreateFrame("EditBox",nil,theFrame, "InputBoxTemplate")
	me.Frame.TextBox:SetWidth(160)
	me.Frame.TextBox:SetHeight(13)
	me.Frame.TextBox:SetPoint("BOTTOMRIGHT",theFrame,"BOTTOMRIGHT",-12,12)
	me.Frame.TextBox:SetAutoFocus(false)

	me.Frame.ScrollBar=CreateFrame("SCROLLFRAME","Protractor_ScrollBar",theFrame,"FauxScrollFrameTemplate")
	me.Frame.ScrollBar:SetScript("OnVerticalScroll", function() FauxScrollFrame_OnVerticalScroll(14, me.UpdateWindow) end)

	me.Frame.ScrollBar:SetPoint("TOPLEFT", me.Frame.Name[1], "TOPLEFT", 0, 0)
	me.Frame.ScrollBar:SetPoint("BOTTOMRIGHT", me.Frame.Percent[10], "BOTTOM", 75, 0)

	me.Frame:Hide()
end


--Setup the variables-
function me:FunctionCPUUsageRecursing(table,key,depth)
	depth=depth or 0
	for k,v in pairs(table) do
		local k2=tostring(k)
		if type(v)=="function" then
			local time, count = GetFunctionCPUUsage(v,false)
			me:AddChartData(key..k2,time,count)
		elseif depth>0 and type(v)=="table" and k2~="_G" then
			me:FunctionCPUUsageRecursing(v,key..k2..".",find,depth-1)
		end
	end
end

function me:FuncCPUUsage()
	if GetCVar("scriptProfile")~="1" then
		Protractor:Print("CPU Profiling is currently disabled")
		return
	end

	local key=me.Frame.TextBox:GetText()
	
	if key=="" then
		key="_G"
	end

	local Remaining
	if string.find(key,"%.") then
		key, Remaining=string.match(key,"(%w*)%.(.*)")
	end
	
	local Last
	local Var=getglobal(key)
	

	
	if Remaining then
		for key, Remaining in string.gmatch(Remaining,"(%w*)%.(.*)") do
			Var=Var[key]
			Last=Remaining
		end
		if Last then
			Var=Var[Last]
		else
			Var=Var[Remaining]
		end
	end

	UpdateAddOnCPUUsage()

	me:ResetChart()
	
	if Var then
		if type(Var)=="function" then
			local time, count = GetFunctionCPUUsage(Var,false)
			me:AddChartData(key,time,count)
		else
			me:FunctionCPUUsageRecursing(Var,"",2)
		end

		me:CompleteChart()
	else
		me:Print("Invalid Variable "..me.Frame.TextBox:GetText())
	end

	me.Frame.Title:SetText("Function CPU Usage")
	me.Frame.UsageTitle:SetText("Time (ms)")
end


--0 = Memory Usage
--1 = CPU Usage
me.Displaying=0	
me.MaxDisplay=3
me.CutoffValue=0.025
me.DisplayFunc={}
me.DisplayFunc[0]=me.MemoryUsage
me.DisplayFunc[1]=me.CPUUsage
me.DisplayFunc[2]=me.FuncCPUUsage

me.DisplayFunc[me.Displaying]()

function me:OnEnable()
	Protractor:Print("Protractor Loaded")
	Protractor:Print("Use /pro or /protractor to show")
	Protractor:Print("Specify a table/function in the edit box for CPU Usage of Functions (example use _G for all functions)")

	
end
